#!/bin/bash
# {{ strategy_name }} - Jupyter Paper Trading Script
# Generated by qubx init

echo "🚀 Starting {{ strategy_class }} in Jupyter paper mode..."
poetry run qubx run config.yml --paper --jupyter