[project]
name = "{{ strategy_name }}"
version = "0.1.0"
description = "{{ strategy_class }} - MACD crossover strategy"
requires-python = ">=3.10"
dependencies = [
    "qubx>=0.6.0",
]

[project.optional-dependencies]
dev = [
    "jupyter",
    "ipykernel",
]

[build-system]
requires = ["setuptools>=61.0"]
build-backend = "setuptools.build_meta"